from transformers import BertTokenizer, BertModel
from ..baseExtractor import baseTextExtractor
import torch
import numpy as np


class bertExtractor(baseTextExtractor):
    """
    Text feature extractor using BERT
    Ref: https://huggingface.co/docs/transformers/model_doc/bert
    Pretrained models: https://huggingface.co/models
    """
    def __init__(self, config, logger):
        try:
            logger.info("Initializing BERT text feature extractor.")
            super().__init__(config, logger)
            self.tokenizer = BertTokenizer.from_pretrained(self.config['pretrained'])
            self.model = BertModel.from_pretrained(self.config['pretrained']).to(self.config['device'])
            self.finetune = self.config['finetune'] if 'finetune' in self.config else False
        except Exception as e:
            logger.error("Failed to initialize bertExtractor.")
            raise e
    
    def extract(self, text):
        try:
            input_ids = self.tokenizer.encode(text, add_special_tokens=True, return_tensors='pt').to(self.config['device'])
            # encoded_inputs = self.tokenizer(text, add_special_tokens=True, return_tensors='pt', padding='max_length', truncation=True, max_length=50).to(self.config['device'])
            with torch.no_grad():
                last_hidden_state = self.model(input_ids).last_hidden_state
                # last_hidden_state = self.model(encoded_inputs['input_ids'], encoded_inputs['attention_mask'], encoded_inputs['token_type_ids']).last_hidden_state
            return last_hidden_state.squeeze().cpu().numpy()
        except Exception as e:
            self.logger.error(f"Failed to extract text features with BERT for '{text}'.")
            raise e

    def tokenize(self, text):
        """
        For compatibility with feature files generated by MMSA DataPre.py

        Returns:
            input_ids: input_ids,
            input_mask: attention_mask,
            segment_ids: token_type_ids
        """
        try:
            # input_ids = self.tokenizer.encode(text, add_special_tokens=True)
            # input_ids = np.expand_dims(input_ids, 1)
            # input_mask = np.ones_like(input_ids)
            # segment_ids = np.zeros_like(input_ids)
            # text_bert = np.concatenate([input_ids, input_mask, segment_ids], axis=1)
            text_bert = self.tokenizer(text, add_special_tokens=True, return_tensors='np')
            text_bert = np.concatenate([text_bert['input_ids'].transpose(), text_bert['attention_mask'].transpose(), text_bert['token_type_ids'].transpose()], axis=1)
            return text_bert
        except Exception as e:
            self.logger.error(f"Failed to tokenize text with BERT for '{text}'.")
            raise e


import librosa


class baseExtractor(object):
    """
    Base class for all extractors.
    """
    def __init__(self, config, logger):
        self.config = config
        self.logger = logger

    def extract(self, file):
        """
        Extract features from input file.
        """
        raise NotImplementedError("extract() not implemented")


class baseAudioExtractor(baseExtractor):
    """
    Base class for all audio extractors.
    """
    def __init__(self, config, logger):
        super().__init__(config, logger)

    def load_audio(self, file):
        """
        Load audio file using librosa.
        """
        y, sr = librosa.load(file, sr=self.config['sample_rate'])
        return y, sr


class baseTextExtractor(baseExtractor):
    """
    Base class for all text extractors.
    """
    def __init__(self, config, logger):
        super().__init__(config, logger)

    def load_text(self, file):
        """
        Load text from file.
        """
        with open(file, 'r') as f:
            text = f.read()
        return text

    def tokenize(self, text):
        """
        Tokenize the input text.
        """
        raise NotImplementedError("tokenize() not implemented")

# class baseVideoExtractor(baseExtractor):
#     """
#     Base class for all video extractors.
#     """
#     def __init__(self, config, logger):
#         super().__init__(config, logger)

#     def load_images(self, img_dir):
#         """
#         Load image files using cv2.
#         """
#         images = []
#         for image_path in sorted(glob(osp.join(img_dir, '*.bmp'))):
#             name = Path(image_path).stem
#             image = cv2.cvtColor(cv2.imread(image_path), cv2.COLOR_BGR2RGB)
#             images.append(image)
#         return images


import torch
from transformers import Wav2Vec2FeatureExtractor, Wav2Vec2Model
from ..baseExtractor import baseAudioExtractor


class wav2vec2Extractor(baseAudioExtractor):
    """
    Audio feature extractor using Wav2Vec2. 
    Ref: https://huggingface.co/transformers/model_doc/wav2vec2.html
    """
    def __init__(self, config, logger):
        try:
            logger.info("Initializing Wav2Vec2 audio feature extractor.")
            super().__init__(config, logger)
            self.preprocessor = Wav2Vec2FeatureExtractor.from_pretrained(config['pretrained'])
            self.extractor = Wav2Vec2Model.from_pretrained(config['pretrained'])
        except Exception as e:
            self.logger.error("Failed to initialize Wav2VecExtractor.")
            raise e

    def extract(self, file):
        """
        Function:
            Extract features from audio file using wav2vec2 pretrained model.

        Parameters:
            file: path to audio file

        Returns:
            audio_result: extracted audio features in numpy array
        """
        try:
            y, sr = self.load_audio(file)
            audio_result = self.preprocessor(y, sampling_rate=sr, return_tensors="pt").input_values
            with torch.no_grad():
                audio_result = self.extractor(audio_result).last_hidden_state.squeeze(0)
            return audio_result
        except Exception as e:
            self.logger.error(f"Failed to extract audio features with Wav2Vec2 from {file}.")
            raise e



import os
import platform
from pathlib import Path

import numpy as np
import pandas as pd

from ..baseExtractor import baseExtractor


class openfaceExtractor(baseExtractor):
    """
    Video feature extractor using OpenFace. 
    Ref: https://mediapipe.dev/
    """
    def __init__(self, config, logger):
        try:
            logger.info("Initializing OpenFace video feature extractor...")
            super().__init__(config, logger)
            if self.config['average_over'] < 1:
                self.pool_size = 1
                logger.warning("'average_over' is less than 1, set to 1.")
            else:
                self.pool_size = self.config['average_over']
                
            self.args = self._parse_args(self.config['args'])
            self.tool_dir = Path(__file__).parent.parent.parent / "exts" / "OpenFace"
            if platform.system() == 'Windows':
                self.tool = self.tool_dir / "FeatureExtraction.exe"
            elif platform.system() == 'Linux':
                self.tool = self.tool_dir / "FeatureExtraction"
            else:
                raise RuntimeError("Cannot Determine OS type.")
            if not self.tool.is_file():
                raise FileNotFoundError("OpenFace tool not found.")
        except Exception as e:
            self.logger.error("Failed to initialize mediapipeExtractor.")
            raise e

    def _parse_args(self, args):
        res = []
        if 'hogalign' in args and args['hogalign']:
            res.append('-hogalign')
        if 'simalign' in args and args['simalign']:
            res.append('-simalign')
        if 'nobadaligned' in args and args['nobadaligned']:
            res.append('-nobadaligned')
        if 'tracked' in args and args['tracked']:
            res.append('-tracked')
        if 'pdmparams' in args and args['pdmparams']:
            res.append('-pdmparams')
        if 'landmark_2D' in args and args['landmark_2D']:
            res.append('-2Dfp')
        if 'landmark_3D' in args and args['landmark_3D']:
            res.append('-3Dfp')
        if 'head_pose' in args and args['head_pose']:
            res.append('-pose')
        if 'action_units' in args and args['action_units']:
            res.append('-aus')
        if 'gaze' in args and args['gaze']:
            res.append('-gaze')
        return res
    
    def extract(self, img_dir, video_name=None, tool_output=False):
        """
        Function:
            Extract features from video file using OpenFace.

        Parameters:
            img_dir: path to directory of images.
            video_name: video name used to save annotation images.
            tool_output: if False, disable stdout of OpenFace tool.

        Returns:
            video_result: extracted video features in numpy array.
        """
        try:
            args = self.args.copy()
            args.extend(['-fdir', img_dir, '-out_dir', img_dir])
            if not tool_output:
                args.append('-quiet')
            cmd = str(self.tool) + " " + " ".join(args)
            os.system(cmd)

            name = Path(img_dir).stem
            df = pd.read_csv(Path(img_dir) / (str(name) + '.csv'))
            features, local_features = [], []
            for i in range(len(df)):
                local_features.append(np.array(df.loc[i][df.columns[5:]]))
                if (i + 1) % self.pool_size == 0:
                    features.append(np.array(local_features).mean(axis=0))
                    local_features = []
            if len(local_features) != 0:
                features.append(np.array(local_features).mean(axis=0))
            return np.array(features)
        except Exception as e:
            self.logger.error(f"Failed to extract video features with OpenFace from {video_name}.")
            raise e



import subprocess
import sys
import time

import pandas
import torch
import torch.nn as nn
import tqdm

from .loss import lossA, lossAV, lossV
from .model import talkNetModel


class talkNet(nn.Module):
    def __init__(self, lr = 0.0001, lrDecay = 0.95, **kwargs):
        super(talkNet, self).__init__()        
        self.model = talkNetModel().cuda()
        self.lossAV = lossAV().cuda()
        self.lossA = lossA().cuda()
        self.lossV = lossV().cuda()
        self.optim = torch.optim.Adam(self.parameters(), lr = lr)
        self.scheduler = torch.optim.lr_scheduler.StepLR(self.optim, step_size = 1, gamma=lrDecay)
        # print(time.strftime("%m-%d %H:%M:%S") + " Model para number = %.2f"%(sum(param.numel() for param in self.model.parameters()) / 1024 / 1024))

    def train_network(self, loader, epoch, **kwargs):
        self.train()
        self.scheduler.step(epoch - 1)
        index, top1, loss = 0, 0, 0
        lr = self.optim.param_groups[0]['lr']        
        for num, (audioFeature, visualFeature, labels) in enumerate(loader, start=1):
            self.zero_grad()
            audioEmbed = self.model.forward_audio_frontend(audioFeature[0].cuda()) # feedForward
            visualEmbed = self.model.forward_visual_frontend(visualFeature[0].cuda())
            audioEmbed, visualEmbed = self.model.forward_cross_attention(audioEmbed, visualEmbed)
            outsAV= self.model.forward_audio_visual_backend(audioEmbed, visualEmbed)  
            outsA = self.model.forward_audio_backend(audioEmbed)
            outsV = self.model.forward_visual_backend(visualEmbed)
            labels = labels[0].reshape((-1)).cuda() # Loss
            nlossAV, _, _, prec = self.lossAV.forward(outsAV, labels)
            nlossA = self.lossA.forward(outsA, labels)
            nlossV = self.lossV.forward(outsV, labels)
            nloss = nlossAV + 0.4 * nlossA + 0.4 * nlossV
            loss += nloss.detach().cpu().numpy()
            top1 += prec
            nloss.backward()
            self.optim.step()
            index += len(labels)
            sys.stderr.write(time.strftime("%m-%d %H:%M:%S") + \
            " [%2d] Lr: %5f, Training: %.2f%%, "    %(epoch, lr, 100 * (num / loader.__len__())) + \
            " Loss: %.5f, ACC: %2.2f%% \r"        %(loss/(num), 100 * (top1/index)))
            sys.stderr.flush()  
        sys.stdout.write("\n")      
        return loss/num, lr

    def evaluate_network(self, loader, evalCsvSave, evalOrig, **kwargs):
        self.eval()
        predScores = []
        for audioFeature, visualFeature, labels in tqdm.tqdm(loader):
            with torch.no_grad():                
                audioEmbed  = self.model.forward_audio_frontend(audioFeature[0].cuda())
                visualEmbed = self.model.forward_visual_frontend(visualFeature[0].cuda())
                audioEmbed, visualEmbed = self.model.forward_cross_attention(audioEmbed, visualEmbed)
                outsAV= self.model.forward_audio_visual_backend(audioEmbed, visualEmbed)  
                labels = labels[0].reshape((-1)).cuda()             
                _, predScore, _, _ = self.lossAV.forward(outsAV, labels)    
                predScore = predScore[:,1].detach().cpu().numpy()
                predScores.extend(predScore)
        evalLines = open(evalOrig).read().splitlines()[1:]
        labels = []
        labels = pandas.Series( ['SPEAKING_AUDIBLE' for line in evalLines])
        scores = pandas.Series(predScores)
        evalRes = pandas.read_csv(evalOrig)
        evalRes['score'] = scores
        evalRes['label'] = labels
        evalRes.drop(['label_id'], axis=1,inplace=True)
        evalRes.drop(['instance_id'], axis=1,inplace=True)
        evalRes.to_csv(evalCsvSave, index=False)
        cmd = "python -O utils/get_ava_active_speaker_performance.py -g %s -p %s "%(evalOrig, evalCsvSave)
        mAP = float(str(subprocess.run(cmd, shell=True, capture_output =True).stdout).split(' ')[2][:5])
        return mAP

    def saveParameters(self, path):
        torch.save(self.state_dict(), path)

    def loadParameters(self, path):
        selfState = self.state_dict()
        loadedState = torch.load(path)
        for name, param in loadedState.items():
            origName = name;
            if name not in selfState:
                name = name.replace("module.", "")
                if name not in selfState:
                    print("%s is not in the model."%origName)
                    continue
            if selfState[name].size() != loadedState[origName].size():
                sys.stderr.write("Wrong parameter length: %s, model: %s, loaded: %s"%(origName, selfState[name].size(), loadedState[origName].size()))
                continue
            selfState[name].copy_(param)



import numpy as np
import librosa
from ..baseExtractor import baseAudioExtractor


class librosaExtractor(baseAudioExtractor):
    """
    Audio feature extractor using librosa. 
    Ref: https://librosa.org/doc/latest/index.html
    """
    def __init__(self, config, logger):
        try:
            logger.info("Initializing librosa audio feature extractor.")
            super().__init__(config, logger)
        except Exception as e:
            self.logger.error("Failed to initialize librosaExtractor.")
            raise e

    def extract(self, file):
        """
        Function:
            Extract features from audio file using librosa.

        Parameters:
            file: path to audio file

        Returns:
            audio_result: extracted audio features in numpy array
        """
        try:
            y, sr = self.load_audio(file)
            res = {}
            for audio_feature in self.config['args'].keys():
                kwargs = self.config['args'][audio_feature]
                method = getattr(librosa.feature, audio_feature)
                try:
                    res[audio_feature] = method(y=y, sr=sr, **kwargs).T
                except TypeError:
                    res[audio_feature] = method(y=y, **kwargs).T
            # concatenate all features
            audio_result = np.concatenate(list(res.values()), axis=1)
            return audio_result
        except Exception as e:
            self.logger.error(f"Failed to extract audio features with librosa from {file}.")
            raise e



import json
import ffmpeg
import os.path as osp
import urllib.request
from tqdm import tqdm
import re
from pathlib import Path


def get_default_config(tool_name: str) -> dict:
    """
    Get default configuration for a tool.

    Args:
        tool_name: name of the tool.

    Returns:
        Python dictionary containing the config.
    """
    path = Path(__file__).parent / "example_configs" / f"{tool_name}.json"
    with open(path, 'r') as f:
        res = json.load(f)
    return res


def get_codec_name(file, mode):
    """
    Function:
        Get video/audio codec of the file.

    Parameters:
        file: Path to the file.
        mode: Should be 'video' or 'audio'.

    Returns:
        codec: Codec name.
        
    """
    assert mode in ['audio', 'video'], "Parameter 'mode' must be 'audio' or 'video'."

    prob_result = ffmpeg.probe(file)
    for track in prob_result['streams']:
        if track['codec_type'] == mode:
            return track['codec_name']



def ffmpeg_extract(in_file, out_path, mode='audio', fps=25):
    """
    Function:
        Extract audio/image from the input file.

    Params:
        in_file: Path to the input file.
        out_path: Path to the output file.
        mode: Should be 'audio' or 'image'.
        fps: Frames per second, will be ignored if mode is 'audio'.

    """
    assert mode in ['audio', 'image'], "Parameter 'mode' must be 'audio' or 'image'."
    
    if mode == 'audio':
        ffmpeg.input(in_file).output(out_path)\
            .run(overwrite_output=True, quiet=True)
    elif mode == 'image':
        ffmpeg.input(in_file)\
            .output(osp.join(out_path, '%03d.bmp'), r=f'{fps}/1')\
            .run(quiet=True)


def download_file(url, save_path):
    """
    Function:
        Download file from url.

    Params:
        url: Url of file to be downloaded.
        save_path: Save path, including filename and extension.

    """
    with urllib.request.urlopen(url) as response, open(save_path, 'wb') as out_file:
        total_size = int(response.info().get('Content-Length', -1))
        if total_size < 0:
            print("Unknown file size")
            data = response.read()
            out_file.write(data)
        else:
            print("Downloading: %s Bytes: %s" % (save_path, total_size))
            pbar = tqdm(total=total_size, unit='iB', unit_scale=True)
            while True:
                data = response.read(1024)
                if not data:
                    break
                out_file.write(data)
                pbar.update(len(data))
            pbar.close()


def atoi(text):
    return int(text) if text.isdigit() else text


def natural_keys(text):
    return [ atoi(c) for c in re.split('(\d+)', text) ]


import torch
import torch.nn as nn

from .audioEncoder      import audioEncoder
from .visualEncoder     import visualFrontend, visualTCN, visualConv1D
from .attentionLayer    import attentionLayer

class talkNetModel(nn.Module):
    def __init__(self):
        super(talkNetModel, self).__init__()
        # Visual Temporal Encoder
        self.visualFrontend  = visualFrontend() # Visual Frontend 
        # self.visualFrontend.load_state_dict(torch.load('visual_frontend.pt', map_location="cuda"))
        # for param in self.visualFrontend.parameters():
        #     param.requires_grad = False       
        self.visualTCN       = visualTCN()      # Visual Temporal Network TCN
        self.visualConv1D    = visualConv1D()   # Visual Temporal Network Conv1d

        # Audio Temporal Encoder 
        self.audioEncoder  = audioEncoder(layers = [3, 4, 6, 3],  num_filters = [16, 32, 64, 128])
        
        # Audio-visual Cross Attention
        self.crossA2V = attentionLayer(d_model = 128, nhead = 8)
        self.crossV2A = attentionLayer(d_model = 128, nhead = 8)

        # Audio-visual Self Attention
        self.selfAV = attentionLayer(d_model = 256, nhead = 8)

    def forward_visual_frontend(self, x):
        B, T, W, H = x.shape  
        x = x.view(B*T, 1, 1, W, H)
        x = (x / 255 - 0.4161) / 0.1688
        x = self.visualFrontend(x)
        x = x.view(B, T, 512)        
        x = x.transpose(1,2)     
        x = self.visualTCN(x)
        x = self.visualConv1D(x)
        x = x.transpose(1,2)
        return x

    def forward_audio_frontend(self, x):    
        x = x.unsqueeze(1).transpose(2, 3)        
        x = self.audioEncoder(x)
        return x

    def forward_cross_attention(self, x1, x2):
        x1_c = self.crossA2V(src = x1, tar = x2)
        x2_c = self.crossV2A(src = x2, tar = x1)        
        return x1_c, x2_c

    def forward_audio_visual_backend(self, x1, x2): 
        x = torch.cat((x1,x2), 2)    
        x = self.selfAV(src = x, tar = x)       
        x = torch.reshape(x, (-1, 256))
        return x    

    def forward_audio_backend(self,x):
        x = torch.reshape(x, (-1, 128))
        return x

    def forward_visual_backend(self,x):
        x = torch.reshape(x, (-1, 128))
        return x




from .audio.librosa import librosaExtractor
from .audio.opensmile import opensmileExtractor
from .audio.wave2vec import wav2vec2Extractor
from .video.mediapipe import mediapipeExtractor
from .video.openface import openfaceExtractor
from .text.bert import bertExtractor
from .text.roberta import robertaExtractor
from .text.glove import gloveExtractor

__all__ = ['AUDIO_EXTRACTOR_MAP', 'VIDEO_EXTRACTOR_MAP', 'TEXT_EXTRACTOR_MAP']

AUDIO_EXTRACTOR_MAP = {
    "librosa": librosaExtractor,
    "opensmile": opensmileExtractor,
    "wav2vec": wav2vec2Extractor,
}

VIDEO_EXTRACTOR_MAP = {
    "mediapipe": mediapipeExtractor,
    "openface": openfaceExtractor,
    # "3dcnn": 3dcnn_extractor,
}

TEXT_EXTRACTOR_MAP = {
    "bert": bertExtractor,
    "roberta": robertaExtractor,
    "glove": gloveExtractor,
}


import platform
import sys
from pathlib import Path
import shutil

import gdown

OPENFACE_BASE = "https://drive.google.com/file/d/17T1WSbkwBokTqNq1IKu51oMX8P122pu-/view?usp=sharing"
OPENFACE_LINUX = "https://drive.google.com/file/d/1Y-WX9xoCYVQqy-VYXyA3697RCiSAldml/view?usp=sharing"
OPENFACE_WIN64 = "https://drive.google.com/file/d/1oPTB5m7K8QPR7o1E6nWMDmSaNkBBF93g/view?usp=sharing"
PRETRAINED = "https://drive.google.com/file/d/1Yzetew3mRjQqhhHLsZP1uT6zT9lR_DOG/view?usp=sharing"

# Check file availability
def integrity_check():
    ext_dir = Path(__file__).parent / "exts"
    if not ext_dir.exists():
        ext_dir.mkdir()
    openface_dir = ext_dir / "OpenFace"
    pretrained_dir = ext_dir / "pretrained"
    res = {}
    res["OpenFace"] = True if openface_dir.exists() else False
    res["pretrained"] = True if pretrained_dir.exists() else False
    return res

# Download from Google Drive and extract
def download_and_extract(url, save_path, remove_zip=True, proxy=None):
    gdown.download(url, save_path, quiet=False, fuzzy=True, proxy=proxy)
    gdown.extractall(save_path)
    if remove_zip:
        Path(save_path).unlink()

# Download OpenFace
def download_openface(proxy=None):
    is_64bits = sys.maxsize > 2**32 # check if system is 64-bit
    if not is_64bits:
        print("We only provide 64-bit version of OpenFace. For 32-bit version, please download the source code and complie it manually.")
        return
    system = platform.system()
    if system == "":
        print("Cannot determine system type. Please download OpenFace manually.")
        return
    ext_dir = Path(__file__).parent / "exts"
    print("Downloading OpenFace...")
    download_and_extract(OPENFACE_BASE, str(ext_dir / "OpenFace_Base.zip"), proxy=proxy)
    # download_and_extract(OPENFACE_BASE, "/home/mhs/OpenFace_Base.zip", proxy=proxy)
    if system == "Linux":
        download_and_extract(OPENFACE_LINUX, str(ext_dir / "OpenFace" / "OpenFace_Linux.zip"), proxy=proxy)
        # download_and_extract(OPENFACE_LINUX, "/home/mhs/OpenFace/OpenFace_Linux.zip", proxy=proxy)
    if system == "Windows":
        download_and_extract(OPENFACE_WIN64, str(ext_dir / "OpenFace" / "OpenFace_Win64.zip"), proxy=proxy)
        # download_and_extract(OPENFACE_WIN64, "/home/mhs/OpenFace_Win64.zip", proxy=proxy)
    print("Done.")

# Download pretrained models
def download_pretrained(proxy=None):
    ext_dir = Path(__file__).parent / "exts"
    download_and_extract(PRETRAINED, str(ext_dir / "pretrained.zip"), proxy=proxy)


def download_missing(proxy=None):
    # For first run. Will only download if the whole folder is missing. 
    res = integrity_check()
    if not res["OpenFace"]:
        download_openface(proxy=proxy)
    if not res["pretrained"]:
        download_pretrained(proxy=proxy)

def force_redownload(proxy=None):
    ext_dir = Path(__file__).parent / "exts"
    ext_dir.mkdir("")
    shutil.rmtree(ext_dir / "OpenFace")
    download_openface(proxy=proxy)
    shutil.rmtree(ext_dir / "pretrained")
    download_pretrained(proxy=proxy)



from transformers import RobertaTokenizer, RobertaModel
from ..baseExtractor import baseTextExtractor
import torch
import numpy as np


class robertaExtractor(baseTextExtractor):
    """
    Text feature extractor using RoBERTa
    Ref: https://huggingface.co/docs/transformers/model_doc/roberta
    Pretrained models: https://huggingface.co/models
    """
    def __init__(self, config, logger):
        try:
            logger.info("Initializing RoBERTa text feature extractor.")
            super().__init__(config, logger)
            self.tokenizer = RobertaTokenizer.from_pretrained(self.config['pretrained'])
            self.model = RobertaModel.from_pretrained(self.config['pretrained']).to(self.config['device'])
            self.finetune = self.config['finetune'] if 'finetune' in self.config else False
        except Exception as e:
            logger.error("Failed to initialize robertaExtractor.")
            raise e
    
    def extract(self, text):
        try:
            input_ids = self.tokenizer.encode(text, add_special_tokens=True, return_tensors='pt').to(self.config['device'])
            with torch.no_grad():
                last_hidden_state = self.model(input_ids).last_hidden_state
            return last_hidden_state.squeeze().cpu().numpy()
        except Exception as e:
            self.logger.error(f"Failed to extract text features with RoBERTa for '{text}'.")
            raise e

    def tokenize(self, text):
        """
        For compatibility with feature files generated by MMSA DataPre.py
        Returns:
            input_ids: input_ids,
            input_mask: attention_mask,
            segment_ids: token_type_ids
        """
        try:
            text_bert = self.tokenizer(text, add_special_tokens=True, return_tensors='np')
            token_type_ids = self.tokenizer.create_token_type_ids_from_sequences(text_bert['input_ids'].squeeze(0))
            text_bert = np.concatenate([text_bert['input_ids'].transpose(), 
                                        text_bert['attention_mask'].transpose(), 
                                        np.expand_dims(token_type_ids, axis=1)],
                                        axis=1)
            return text_bert
        except Exception as e:
            self.logger.error(f"Failed to tokenize text with RoBERTa for '{text}'.")
            raise e


import glob
import math
import os
import subprocess
import warnings
from pathlib import Path
from shutil import move, rmtree

import cv2
import numpy
import python_speech_features
import torch
from scenedetect.detectors import ContentDetector
from scenedetect.scene_manager import SceneManager
from scenedetect.stats_manager import StatsManager
from scenedetect.video_manager import VideoManager
from scipy import signal
from scipy.interpolate import interp1d
from scipy.io import wavfile

from .model import S3FD
from .talkNet import talkNet

warnings.filterwarnings("ignore")

def scene_detect(args):
    # CPU: Scene detection, output is the list of each shot's time duration
    videoManager = VideoManager([args['videoFilePath']])
    statsManager = StatsManager()
    sceneManager = SceneManager(statsManager)
    sceneManager.add_detector(ContentDetector())
    baseTimecode = videoManager.get_base_timecode()
    videoManager.set_downscale_factor()
    videoManager.start()
    sceneManager.detect_scenes(frame_source = videoManager, show_progress=False)
    sceneList = sceneManager.get_scene_list(baseTimecode)
    if sceneList == []:
        sceneList = [(videoManager.get_base_timecode(),videoManager.get_current_timecode())]
    return sceneList

def inference_video(args):
    # GPU: Face detection, output is the list contains the face location and score in this frame
    DET = S3FD(device='cuda')
    flist = glob.glob(os.path.join(args['pyframesPath'], '*.jpg'))
    flist.sort()
    dets = []
    for fidx, fname in enumerate(flist):
        image = cv2.imread(fname)
        imageNumpy = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        bboxes = DET.detect_faces(imageNumpy, conf_th=0.9, scales=[args['facedetScale']])
        dets.append([])
        for bbox in bboxes:
          dets[-1].append({'frame':fidx, 'bbox':(bbox[:-1]).tolist(), 'conf':bbox[-1]}) # dets has the frames info, bbox info, conf info
        # sys.stderr.write('%s-%05d; %d dets\r' % (args['videoFilePath'], fidx, len(dets[-1])))
    return dets

def bb_intersection_over_union(boxA, boxB, evalCol = False):
    # CPU: IOU Function to calculate overlap between two image
    xA = max(boxA[0], boxB[0])
    yA = max(boxA[1], boxB[1])
    xB = min(boxA[2], boxB[2])
    yB = min(boxA[3], boxB[3])
    interArea = max(0, xB - xA) * max(0, yB - yA)
    boxAArea = (boxA[2] - boxA[0]) * (boxA[3] - boxA[1])
    boxBArea = (boxB[2] - boxB[0]) * (boxB[3] - boxB[1])
    if evalCol == True:
        iou = interArea / float(boxAArea)
    else:
        iou = interArea / float(boxAArea + boxBArea - interArea)
    return iou

def track_shot(args, sceneFaces):
    # CPU: Face tracking
    iouThres  = 0.5     # Minimum IOU between consecutive face detections
    tracks    = []
    while True:
        track     = []
        for frameFaces in sceneFaces:
            for face in frameFaces:
                if track == []:
                    track.append(face)
                    frameFaces.remove(face)
                elif face['frame'] - track[-1]['frame'] <= args['numFailedDet']:
                    iou = bb_intersection_over_union(face['bbox'], track[-1]['bbox'])
                    if iou > iouThres:
                        track.append(face)
                        frameFaces.remove(face)
                        continue
                else:
                    break
        if track == []:
            break
        elif len(track) > args['minTrack']:
            frameNum    = numpy.array([ f['frame'] for f in track ])
            bboxes      = numpy.array([numpy.array(f['bbox']) for f in track])
            frameI      = numpy.arange(frameNum[0],frameNum[-1]+1)
            bboxesI    = []
            for ij in range(0,4):
                interpfn  = interp1d(frameNum, bboxes[:,ij])
                bboxesI.append(interpfn(frameI))
            bboxesI  = numpy.stack(bboxesI, axis=1)
            if max(numpy.mean(bboxesI[:,2]-bboxesI[:,0]), numpy.mean(bboxesI[:,3]-bboxesI[:,1])) > args['minFaceSize']:
                tracks.append({'frame':frameI,'bbox':bboxesI})
    return tracks

def crop_video(args, track, cropFile):
    # CPU: crop the face clips
    flist = glob.glob(os.path.join(args['pyframesPath'], '*.jpg')) # Read the frames
    flist.sort()
    vOut = cv2.VideoWriter(cropFile + 't.avi', cv2.VideoWriter_fourcc(*'XVID'), 25, (224,224))# Write video
    dets = {'x':[], 'y':[], 's':[]}
    for det in track['bbox']: # Read the tracks
        dets['s'].append(max((det[3]-det[1]), (det[2]-det[0]))/2) 
        dets['y'].append((det[1]+det[3])/2) # crop center x 
        dets['x'].append((det[0]+det[2])/2) # crop center y
    dets['s'] = signal.medfilt(dets['s'], kernel_size=13)  # Smooth detections 
    dets['x'] = signal.medfilt(dets['x'], kernel_size=13)
    dets['y'] = signal.medfilt(dets['y'], kernel_size=13)
    for fidx, frame in enumerate(track['frame']):
        cs  = args['cropScale']
        bs  = dets['s'][fidx]   # Detection box size
        bsi = int(bs * (1 + 2 * cs))  # Pad videos by this amount 
        image = cv2.imread(flist[frame])
        frame = numpy.pad(image, ((bsi,bsi), (bsi,bsi), (0, 0)), 'constant', constant_values=(110, 110))
        my  = dets['y'][fidx] + bsi  # BBox center Y
        mx  = dets['x'][fidx] + bsi  # BBox center X
        face = frame[int(my-bs):int(my+bs*(1+2*cs)),int(mx-bs*(1+cs)):int(mx+bs*(1+cs))]
        vOut.write(cv2.resize(face, (224, 224)))
    audioTmp    = cropFile + '.wav'
    audioStart  = (track['frame'][0]) / 25
    audioEnd    = (track['frame'][-1]+1) / 25
    vOut.release()
    command = ("ffmpeg -y -i %s -async 1 -ac 1 -vn -acodec pcm_s16le -ar 16000 -ss %.3f -to %.3f %s -loglevel panic" % \
              (args['audioFilePath'], audioStart, audioEnd, audioTmp)) 
    subprocess.call(command, shell=True, stdout=None) # Crop audio file
    # _, audio = wavfile.read(audioTmp)
    command = ("ffmpeg -y -i %st.avi -i %s -c:v copy -c:a copy %s.avi -loglevel panic" % \
              (cropFile, audioTmp, cropFile)) # Combine audio and video file
    subprocess.call(command, shell=True, stdout=None)
    os.remove(cropFile + 't.avi')
    return {'track':track, 'proc_track':dets}

def evaluate_network(files, args):
    # GPU: active speaker detection by pretrained TalkNet
    s = talkNet()
    weight_file = Path(__file__).parent.parent / 'exts' / 'pretrained' / 'TalkSet.pth'
    s.loadParameters(weight_file)
    s.eval()
    allScores = []
    # durationSet = {1,2,4,6} # To make the result more reliable
    durationSet = {1,1,1,2,2,2,3,3,4,5,6} # Use this line can get more reliable result
    for file in files:
        fileName = os.path.splitext(file.split('/')[-1])[0] # Load audio and video
        _, audio = wavfile.read(os.path.join(args['pycropPath'], fileName + '.wav'))
        audioFeature = python_speech_features.mfcc(audio, 16000, numcep = 13, winlen = 0.025, winstep = 0.010)
        video = cv2.VideoCapture(os.path.join(args['pycropPath'], fileName + '.avi'))
        videoFeature = []
        while video.isOpened():
            ret, frames = video.read()
            if ret == True:
                face = cv2.cvtColor(frames, cv2.COLOR_BGR2GRAY)
                face = cv2.resize(face, (224,224))
                face = face[int(112-(112/2)):int(112+(112/2)), int(112-(112/2)):int(112+(112/2))]
                videoFeature.append(face)
            else:
                break
        video.release()
        videoFeature = numpy.array(videoFeature)
        length = min((audioFeature.shape[0] - audioFeature.shape[0] % 4) / 100, videoFeature.shape[0])
        audioFeature = audioFeature[:int(round(length * 100)),:]
        videoFeature = videoFeature[:int(round(length * 25)),:,:]
        allScore = [] # Evaluation use TalkNet
        for duration in durationSet:
            batchSize = int(math.ceil(length / duration))
            scores = []
            with torch.no_grad():
                for i in range(batchSize):
                    inputA = torch.FloatTensor(audioFeature[i * duration * 100:(i+1) * duration * 100,:]).unsqueeze(0).cuda()
                    inputV = torch.FloatTensor(videoFeature[i * duration * 25: (i+1) * duration * 25,:,:]).unsqueeze(0).cuda()
                    embedA = s.model.forward_audio_frontend(inputA)
                    embedV = s.model.forward_visual_frontend(inputV)	
                    embedA, embedV = s.model.forward_cross_attention(embedA, embedV)
                    out = s.model.forward_audio_visual_backend(embedA, embedV)
                    score = s.lossAV.forward(out, labels = None)
                    scores.extend(score)
            allScore.append(scores)
        allScore = numpy.round((numpy.mean(numpy.array(allScore), axis = 0)), 1).astype(float)
        allScores.append(allScore)	
    return allScores

def visualization(tracks, scores, args):
    # CPU: visulize the result for video format
    flist = glob.glob(os.path.join(args['pyframesPath'], '*.jpg'))
    flist.sort()
    faces = [[] for i in range(len(flist))]
    for tidx, track in enumerate(tracks):
        score = scores[tidx]
        for fidx, frame in enumerate(track['track']['frame'].tolist()):
            s = score[max(fidx - 2, 0): min(fidx + 3, len(score) - 1)] # average smoothing
            s = numpy.mean(s)
            faces[frame].append({'track':tidx, 'score':float(s),'s':track['proc_track']['s'][fidx], 'x':track['proc_track']['x'][fidx], 'y':track['proc_track']['y'][fidx]})
    talker_time = 0
    cs = args['cropScale']
    for fidx, fname in enumerate(flist):
        image = cv2.imread(fname)
        best_score, best_id = 0, -1
        for fid, face in enumerate(faces[fidx]):
            if face['score'] > best_score:
                best_score = face['score']
                best_id = fid
        if best_id != -1:
            face = faces[fidx][best_id]
            y1 = int(face['y']-face['s']*(1+cs))
            y2 = int(face['y']+face['s']*(1+2*cs))
            x1 = int(face['x']-face['s']*(1+cs))
            x2 = int(face['x']+face['s']*(1+cs))
            if y1 < 0:
                y1 = 0
            if y2 > image.shape[0]:
                y2 = image.shape[0]
            if x1 < 0:
                x1 = 0
            if x2 > image.shape[1]:
                x2 = image.shape[1]
            image_talk = image[y1:y2, x1:x2]
            temp_path = os.path.join(args['pytalkerPath'], str(talker_time).zfill(5) +'.jpg')
            cv2.imwrite(temp_path, image_talk)
            talker_time += 1

def run_ASD(input_video, output_dir, fps, args):

    assert Path(input_video).is_file(), f"{input_video} is not a file"

    # Initialization 
    args['pyaviPath'] = os.path.join(output_dir, 'pyavi')
    args['pyframesPath'] = os.path.join(output_dir, 'pyframes')
    args['pycropPath'] = os.path.join(output_dir, 'pycrop')
    args['pytalkerPath'] = os.path.join(output_dir, 'talker_face')
    os.makedirs(args['pyaviPath'], exist_ok = True) # The path for the input video, input audio, output video
    os.makedirs(args['pyframesPath'], exist_ok = True) # Save all the video frames
    os.makedirs(args['pycropPath'], exist_ok = True) # Save the detected face clips (audio+video) in this process
    os.makedirs(args['pytalkerPath'], exist_ok = True) # Save the detected face clips (audio+video) in this process

    # Extract video
    args['videoFilePath'] = os.path.join(args['pyaviPath'], 'video.avi')
    command = ("ffmpeg -y -i %s -qscale:v 2 -async 1 -r %d %s -loglevel panic" % \
        (input_video, fps, args['videoFilePath']))
    subprocess.call(command, shell=True, stdout=None)
    # sys.stderr.write(time.strftime("%Y-%m-%d %H:%M:%S") + " Extract the video and save in %s \r\n" %(args['videoFilePath']))
    
    # Extract audio
    args['audioFilePath'] = os.path.join(args['pyaviPath'], 'audio.wav')
    command = ("ffmpeg -y -i %s -qscale:a 0 -ac 1 -vn -ar 16000 %s -loglevel panic" % \
        (args['videoFilePath'], args['audioFilePath']))
    subprocess.call(command, shell=True, stdout=None)
    # sys.stderr.write(time.strftime("%Y-%m-%d %H:%M:%S") + " Extract the audio and save in %s \r\n" %(args['audioFilePath']))

    # Extract the video frames
    command = ("ffmpeg -y -i %s -qscale:v 2 -f image2 %s -loglevel panic" % \
        (args['videoFilePath'], os.path.join(args['pyframesPath'], '%06d.jpg'))) 
    subprocess.call(command, shell=True, stdout=None)
    # sys.stderr.write(time.strftime("%Y-%m-%d %H:%M:%S") + " Extract the frames and save in %s \r\n" %(args['pyframesPath']))

    # Scene detection for the video frames
    scene = scene_detect(args)

    # Face detection for the video frames
    faces = inference_video(args)

    # Face tracking
    allTracks, vidTracks = [], []
    for shot in scene:
        if shot[1].frame_num - shot[0].frame_num >= args['minTrack']: # Discard the shot frames less than minTrack frames
            allTracks.extend(track_shot(args, faces[shot[0].frame_num:shot[1].frame_num])) # 'frames' to present this tracks' timestep, 'bbox' presents the location of the faces
    # sys.stderr.write(time.strftime("%Y-%m-%d %H:%M:%S") + " Face track and detected %d tracks \r\n" %len(allTracks))

    # Face clips cropping
    for ii, track in enumerate(allTracks):
        vidTracks.append(crop_video(args, track, os.path.join(args['pycropPath'], '%05d'%ii)))
    # sys.stderr.write(time.strftime("%Y-%m-%d %H:%M:%S") + " Face Crop and saved in %s tracks \r\n" %args['pycropPath'])

    # Active Speaker Detection by TalkNet
    files = glob.glob("%s/*.avi"%args['pycropPath'])
    files.sort()
    scores = evaluate_network(files, args)

    # Visualization, save the result as the new video	
    visualization(vidTracks, scores, args)

    files = glob.glob(os.path.join(args['pytalkerPath'], '*.jpg'))
    for file in files:
        move(file, output_dir)
    rmtree(args['pyaviPath'])
    rmtree(args['pyframesPath'])
    rmtree(args['pycropPath'])
    rmtree(args['pytalkerPath'])


    


import os
import os.path as osp
from glob import glob
from pathlib import Path

import cv2
import numpy as np
from ..baseExtractor import baseExtractor

import mediapipe as mp


class mediapipeExtractor(baseExtractor):
    """
    Video feature extractor using MediaPipe. 
    Ref: https://mediapipe.dev/
    """
    def __init__(self, config, logger):
        try:
            logger.info("Initializing MediaPipe video feature extractor...")
            super().__init__(config, logger)
            self.args = self.config['args']
            if self.args['visualize']: # drawing utilities
                self.mp_drawing = mp.solutions.drawing_utils
                self.mp_drawing_styles = mp.solutions.drawing_styles
                self.drawing_spec = self.mp_drawing.DrawingSpec(
                    thickness=1, circle_radius=1
                )
            if 'holistic' in self.args:
                self.kwargs = self.args['holistic']
                self.method = mp.solutions.holistic.Holistic
            elif 'face_mesh' in self.args:
                self.kwargs = self.args['face_mesh']
                self.kwargs['max_num_faces'] = 1
                self.method = mp.solutions.face_mesh.FaceMesh
        except Exception as e:
            self.logger.error("Failed to initialize mediapipeExtractor.")
            raise e

    def extract(self, img_dir, video_name=None):
        """
        Function:
            Extract features from video file using MediaPipe.

        Parameters:
            img_dir: path to directory of images.
            video_name: video name used to save annotation images.

        Returns:
            video_result: extracted video features in numpy array.
        """
        try:
            video_result = []
            with self.method(static_image_mode=False, **self.kwargs) as method:
                for image_path in sorted(glob(osp.join(img_dir, '*.bmp'))):
                    name = Path(image_path).stem
                    image = cv2.cvtColor(cv2.imread(image_path), cv2.COLOR_BGR2RGB)
                    result = method.process(image)
                    if 'holistic' in self.args:
                        if not result.face_landmarks:
                            self.logger.debug(f"No face detected in {image_path}.")
                            continue
                        if self.args['visualize']:
                            assert video_name is not None, \
                                "video_name should be passed in order to save annotation images."
                            annotated_image = image.copy()
                            condition = np.stack((result.segmentation_mask,) * 3, axis=-1) > 0.1
                            bg_image = np.zeros(image.shape, dtype=np.uint8)
                            annotated_image = np.where(condition, annotated_image, bg_image)
                            self.mp_drawing.draw_landmarks(
                                image=annotated_image,
                                landmark_list=result.face_landmarks,
                                connections=mp.solutions.holistic.FACEMESH_TESSELATION,
                                landmark_drawing_spec = None,
                                connection_drawing_spec = self.mp_drawing_styles
                                .get_default_face_mesh_tesselation_style()
                            )
                            self.mp_drawing.draw_landmarks(
                                image=annotated_image,
                                landmark_list=result.pose_landmarks,
                                connections=mp.solutions.holistic.POSE_CONNECTIONS,
                                landmark_drawing_spec = self.mp_drawing_styles.
                                get_default_pose_landmarks_style()
                            )
                            self.mp_drawing.draw_landmarks(
                                image=annotated_image,
                                landmark_list=result.left_hand_landmarks,
                                connections=mp.solutions.holistic.HAND_CONNECTIONS,
                                landmark_drawing_spec = self.mp_drawing_styles.
                                get_default_hand_landmarks_style()
                            )
                            self.mp_drawing.draw_landmarks(
                                image=annotated_image,
                                landmark_list=result.right_hand_landmarks,
                                connections=mp.solutions.holistic.HAND_CONNECTIONS,
                                landmark_drawing_spec = self.mp_drawing_styles.
                                get_default_hand_landmarks_style()
                            )
                            os.makedirs(osp.join(self.args['visualize_dir'], video_name), exist_ok=True)
                            cv2.imwrite(osp.join(self.args['visualize_dir'], video_name, name + '.jpg'),
                                        cv2.cvtColor(annotated_image, cv2.COLOR_RGB2BGR))
                        
                        res_image = []
                        for landmark in result.face_landmarks.landmark:
                            res_image.append(landmark.x)
                            res_image.append(landmark.y)
                            res_image.append(landmark.z)
                        if result.pose_landmarks:
                            for landmark in result.pose_landmarks.landmark:
                                res_image.append(landmark.x)
                                res_image.append(landmark.y)
                                res_image.append(landmark.z)
                        else:
                            res_image.extend([0]*33*3)
                        if result.left_hand_landmarks:
                            for landmark in result.left_hand_landmarks.landmark:
                                res_image.append(landmark.x)
                                res_image.append(landmark.y)
                                res_image.append(landmark.z)
                        else:
                            res_image.extend([0]*21*3)
                        if result.right_hand_landmarks:
                            for landmark in result.right_hand_landmarks.landmark:
                                res_image.append(landmark.x)
                                res_image.append(landmark.y)
                                res_image.append(landmark.z)
                        else:
                            res_image.extend([0]*21*3)
                        video_result.append(res_image)
                    elif 'face_mesh' in self.args:
                        if not result.multi_face_landmarks:
                            self.logger.debug(f"No face detected in {image_path}.")
                            continue
                        if self.args['visualize']:
                            assert video_name is not None, \
                                "video_name should be passed in order to save annotation images."
                            annotated_image = image.copy()
                            self.mp_drawing.draw_landmarks(
                                image=annotated_image,
                                landmark_list=result.multi_face_landmarks[0],
                                connections=mp.solutions.face_mesh.FACEMESH_TESSELATION,
                                landmark_drawing_spec=None,
                                connection_drawing_spec=self.mp_drawing_styles
                                .get_default_face_mesh_tesselation_style()
                            )
                            self.mp_drawing.draw_landmarks(
                                image=annotated_image,
                                landmark_list=result.multi_face_landmarks[0],
                                connections=mp.solutions.face_mesh.FACEMESH_CONTOURS,
                                landmark_drawing_spec=None,
                                connection_drawing_spec=self.mp_drawing_styles
                                .get_default_face_mesh_contours_style()
                            )
                            os.makedirs(osp.join(self.args['visualize_dir'], video_name), exist_ok=True)
                            cv2.imwrite(osp.join(self.args['visualize_dir'], video_name, name + '.jpg'),
                                        cv2.cvtColor(annotated_image, cv2.COLOR_RGB2BGR))

                        res_image = []
                        for landmark in result.multi_face_landmarks[0].landmark:
                            res_image.append(landmark.x)
                            res_image.append(landmark.y)
                            res_image.append(landmark.z)
                        video_result.append(res_image)
                video_result = np.array(video_result)
                return video_result

        except Exception as e:
            self.logger.error(f"Failed to extract video features with MediaPipe from {video_name}.")
            raise e



from glob import glob
import logging
import os
import os.path as osp
import shutil

import numpy as np
from torch.utils.data import Dataset

from .ASD import run_ASD
from .extractors import *
from .utils import ffmpeg_extract


class FET_Dataset(Dataset):
    """
    Dataset for MMSA-FET
    """

    def __init__(
        self, 
        df, 
        dataset_dir, 
        dataset_name, 
        config,
        dataset_config, 
        tmp_dir,
        ignore_error=True
    ):
        self.df = df
        self.dataset_dir = dataset_dir
        self.dataset_name = dataset_name
        self.config = config
        self.dataset_config = dataset_config
        self.tmp_dir = tmp_dir
        self.ignore_error = ignore_error
        self.annotation_dict = {
            'Negative': -1,
            'Neutral': 0,
            'Positive': 1
        }
        if self.dataset_config:
            if 'annotations' in self.dataset_config and type(self.dataset_config['annotations']) == dict:
                self.annotation_dict = self.dataset_config['annotations']
        self.logger = logging.getLogger("MMSA-FET")
        self.__init_extractors()

    def __len__(self):
        return len(self.df)

    def __init_extractors(self):
        if 'audio' in self.config:
            audio_cfg = self.config['audio']
            extractor_name = audio_cfg['tool']
            self.audio_extractor = AUDIO_EXTRACTOR_MAP[extractor_name](audio_cfg, self.logger)
        if 'video' in self.config:
            video_cfg = self.config['video']
            extractor_name = video_cfg['tool']
            self.video_extractor = VIDEO_EXTRACTOR_MAP[extractor_name](video_cfg, self.logger)
        if 'text' in self.config:
            text_cfg = self.config['text']
            extractor_name = text_cfg['model']
            self.text_extractor = TEXT_EXTRACTOR_MAP[extractor_name](text_cfg, self.logger)

    def __extract_video(self, video_path, video_id):
        # extract images from video
        fps = self.config['video']['fps']
        out_path = osp.join(self.tmp_dir, video_id)
        os.makedirs(out_path, exist_ok=False)

        if 'multiFace' in self.config['video'] and self.config['video']['multiFace']['enable'] == True:
            # enable Active Speaker Detection
            run_ASD(video_path, out_path, fps, self.config['video']['multiFace'])
            if len(glob(osp.join(out_path, '*.jpg'))) == 0:
                self.logger.warning(f'ASD returned empty results for video {video_id}')
                shutil.rmtree(out_path)
                return np.zeros((1,1))
        else:
            ffmpeg_extract(video_path, out_path, mode='image', fps=fps)

        # extract video features
        video_result = self.video_extractor.extract(out_path, video_id)
        # delete tmp images
        # for image_path in glob(osp.join(out_path, '*.bmp')):
        #     os.remove(image_path)
        shutil.rmtree(out_path)
        return video_result

    def __extract_audio(self, video_path, video_id):
        # extract audio from video file
        tmp_audio_file = osp.join(self.tmp_dir, video_id + '.wav')
        ffmpeg_extract(video_path, tmp_audio_file, mode='audio')

        # extract audio features
        audio_result = self.audio_extractor.extract(tmp_audio_file)
        # delete tmp audio file
        os.remove(tmp_audio_file)
        return audio_result

    def __extract_text(self, text):
        # extract text features
        text_result = self.text_extractor.extract(text)
        return text_result

    def __preprocess_text(self, text):
        # tokenize text, for models that use bert
        token_result = self.text_extractor.tokenize(text)
        return token_result

    def __getitem__(self, index):
        video_id, clip_id, text, label, label_T, label_A, label_V, annotation, mode = \
            self.df.iloc[index]['video_id'], self.df.iloc[index]['clip_id'], self.df.iloc[index]['text'], \
            self.df.iloc[index]['label'], self.df.iloc[index]['label_T'], self.df.iloc[index]['label_A'], \
            self.df.iloc[index]['label_V'], self.df.iloc[index]['annotation'], self.df.iloc[index]['mode']
        cur_id = video_id + '$_$' + clip_id
        tmp_id = video_id + '_' + clip_id
        res = {
            'id': cur_id,
            # 'audio': feature_A,
            # 'vision': feature_V,
            'raw_text': text,
            # 'text': feature_T,
            # 'text_bert': text_bert,
            # 'audio_lengths': seq_A,
            # 'vision_lengths': seq_V,
            'annotations': annotation,
            # 'classification_labels': np.nan if np.isnan(annotation) else self.annotation_dict[annotation],
            'regression_labels': label,
            'regression_labels_A': label_A,
            'regression_labels_V': label_V,
            'regression_labels_T': label_T,
            'mode': mode
        }
        # video
        video_path = osp.join(self.dataset_dir, 'Raw', video_id, clip_id + '.mp4')
        try:
            if 'video' in self.config:
                feature_V = self.__extract_video(video_path, tmp_id)
                seq_V = feature_V.shape[0]
                res['vision'] = feature_V
                res['vision_lengths'] = seq_V
            # audio
            if 'audio' in self.config:
                feature_A = self.__extract_audio(video_path, tmp_id)
                seq_A = feature_A.shape[0]
                res['audio'] = feature_A
                res['audio_lengths'] = seq_A
            # text
            if 'text' in self.config:
                feature_T = self.__extract_text(text)
                seq_T = feature_T.shape[0]
                text_bert = self.__preprocess_text(text)
                res['text'] = feature_T
                res['text_bert'] = text_bert
                if type(res['text_bert']) != np.ndarray:
                    res.pop('text_bert')
            return res
        except Exception as e:
            self.logger.error(f'Error occurred when extracting features for video {video_id} clip {clip_id}')
            if self.ignore_error:
                self.logger.warning(f'Ignore error and continue')
                return None
            else:
                raise e
        


import torch
import torch.nn as nn
from torch.nn import functional as F
from torch.nn import MultiheadAttention

class attentionLayer(nn.Module):

    def __init__(self, d_model, nhead, dropout=0.1):
        super(attentionLayer, self).__init__()
        self.self_attn = MultiheadAttention(d_model, nhead, dropout=dropout)
        
        self.linear1 = nn.Linear(d_model, d_model * 4)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(d_model * 4, d_model)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.activation = F.relu

    def forward(self, src, tar):
        # type: (Tensor, Optional[Tensor], Optional[Tensor]) -> Tensor
        src = src.transpose(0, 1) # B, T, C -> T, B, C
        tar = tar.transpose(0, 1) # B, T, C -> T, B, C
        src2 = self.self_attn(tar, src, src, attn_mask=None,
                              key_padding_mask=None)[0]
        src = src + self.dropout1(src2)
        src = self.norm1(src)

        src2 = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = src + self.dropout2(src2)
        src = self.norm2(src)
        src = src.transpose(0, 1) # T, B, C -> B, T, C
        return src



import os
import os.path as osp
from glob import glob
from pathlib import Path

import cv2
import torch
import torch.nn as nn

from ...models import Senet50_ft_dag
from ...utils import download_file
from ..baseExtractor import baseAudioExtractor, baseExtractor


class vggfaceExtractor(baseAudioExtractor):
    """
    Video feature extractor using pretrained model from VGGFace2. 
    Ref: https://www.robots.ox.ac.uk/~albanie/pytorch-models.html
    """
    def __init__(self, config, logger):
        try:
            logger.info("Initializing VGGFace2 video extractor...")
            super().__init__(config, logger)
            self.model = Senet50_ft_dag()
            save_path = Path.home() / '.MSA-FET' / 'pretrained'
            Path(save_path).mkdir(parents=True, exist_ok=True)
            url = "http://www.robots.ox.ac.uk/~albanie/models/pytorch-mcn/senet50_ft_dag.pth"
            if not osp.isfile(osp.join(save_path, 'senet50_ft_dag.pth')):
                download_file(url, save_path / 'senet50_ft_dag.pth')
            self.model.load_state_dict(torch.load(save_path / 'senet50_ft_dag.pth'))
            self.model.eval()
        except Exception as e:
            self.logger.error("Failed to initialize vggfaceExtractor.")
            os.remove(osp.join(save_path, 'senet50_ft_dag.pth'))
            raise e

    def extract(self, img_dir, video_name=None):
        """
        Function:
            Extract features from video file using VGGFace2 pretrained model.

        Parameters:
            img_dir: path to directory of images.

        Returns:
            video_result: extracted video features in numpy array.
        """
        try:
            for image_path in sorted(glob(osp.join(img_dir, '*.bmp'))):
                name = Path(image_path).stem
                image = cv2.cvtColor(cv2.imread(image_path), cv2.COLOR_BGR2RGB)
                image = torch.from_numpy(image)
                self.model(image)
        except Exception as e:
            self.logger.error("Failed to extract video features with VGGFace2 pretrained model.")
            raise e



import torch
import torch.nn as nn
import torch.nn.functional as F

class lossAV(nn.Module):
	def __init__(self):
		super(lossAV, self).__init__()
		self.criterion = nn.CrossEntropyLoss()
		self.FC        = nn.Linear(256, 2)
		
	def forward(self, x, labels=None):	
		x = x.squeeze(1)
		x = self.FC(x)
		if labels == None:
			predScore = x[:,1]
			predScore = predScore.t()
			predScore = predScore.view(-1).detach().cpu().numpy()
			return predScore
		else:
			nloss = self.criterion(x, labels)
			predScore = F.softmax(x, dim = -1)
			predLabel = torch.round(F.softmax(x, dim = -1))[:,1]
			correctNum = (predLabel == labels).sum().float()
			return nloss, predScore, predLabel, correctNum

class lossA(nn.Module):
	def __init__(self):
		super(lossA, self).__init__()
		self.criterion = nn.CrossEntropyLoss()
		self.FC        = nn.Linear(128, 2)

	def forward(self, x, labels):	
		x = x.squeeze(1)
		x = self.FC(x)	
		nloss = self.criterion(x, labels)
		return nloss

class lossV(nn.Module):
	def __init__(self):
		super(lossV, self).__init__()

		self.criterion = nn.CrossEntropyLoss()
		self.FC        = nn.Linear(128, 2)

	def forward(self, x, labels):	
		x = x.squeeze(1)
		x = self.FC(x)
		nloss = self.criterion(x, labels)
		return nloss




import argparse
from pathlib import Path

from .main import FeatureExtractionTool
from .install import download_missing, force_redownload


def parse_args():
    # main parser
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="command")
    parser.add_argument('-v', '--verbose', action='store_true',
                        help="Print more information to stdout.")
    parser.add_argument('-q', '--quiet', action='store_true',
                        help="Print only errors to stdout.")
    parser.add_argument('--tmp-dir', type=str, default=Path.home() / '.MMSA-FET/tmp',
                        help="Temporary directory for storing intermediate results. Default: '~/.MSA-FET/tmp'")
    parser.add_argument('--log-dir', type=str, default=Path.home() / '.MMSA-FET/log',
                        help="Log file directory. Default: '~/.MSA-FET/log'")

    # sub-parser for installation
    parser_install = subparsers.add_parser("install", help="Download models required to run MSA-FET")
    parser_install.add_argument('-p', '--proxy', type=str, default=None,
                                help="Proxy for downloading. e.g. socks5://127.0.0.1:8080")
    parser_install.add_argument('-f', '--force', action='store_true', help="Force re-download models.")
    
    # sub-parser for dataset mode
    parser_single = subparsers.add_parser("run_single", help="Run feature extraction on a single video")
    parser_single.add_argument('-i', '--input', type=str, required=True,
                                help="Input video file in file mode, or dataset dir in dataset mode.")
    parser_single.add_argument('-c', '--config-file', type=str, required=True,
                                help="Path to config file.")
    parser_single.add_argument('-o', '--output', type=str, required=True,
                                help="Path to output pkl file.")
    parser_single.add_argument('-t', '--text-file', type=str, default=None,
                                help="File containing transcriptions of the video. Required when extracting text features.")
    parser_single.add_argument('--return_type', type=str, default='np', choices=['np', 'pt', 'df'],
                                help="Return type of the tool.")
    

    parser_dataset = subparsers.add_parser("run_dataset", help="Run feature extraction on a dataset")
    parser_dataset.add_argument('-i', '--input', type=str, required=True,
                                help="Input dataset dir.")
    parser_dataset.add_argument('-c', '--config-file', type=str, required=True,
                                help="Path to config file.")
    parser_dataset.add_argument('-o', '--output', type=str, required=True,
                                help="Path to output pkl file.")
    parser_dataset.add_argument('-n', '--num-workers', type=int, default=4,
                                help="Num of dataloader workers.")
    parser_dataset.add_argument('--batch-size', type=int, default=4,
                                help="Batch size. Default: 4")
    parser_dataset.add_argument('--return_type', type=str, default='np', choices=['np', 'pt', 'df'],
                                help="Return type of the tool.")

    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()

    if args.command == 'install':
        if args.force:
            force_redownload(args.proxy)
        else:
            download_missing(args.proxy)
    elif args.command is None:
        print("Please specify a command. Use -h to see all available commands.")
    else:
        verbose = 1
        if args.verbose:
            verbose = 2
        if args.quiet:
            verbose = 0
        
        fet = FeatureExtractionTool(
            config=args.config_file,
            tmp_dir=args.tmp_dir,
            log_dir=args.log_dir,
            verbose=verbose
        )

        if args.command == 'run_single':
            fet.run_single(
                in_file=args.input,
                out_file=args.output,
                text_file=args.text_file,
                return_type=args.return_type
            )
        elif args.command == 'run_dataset':
            fet.run_dataset(
                dataset_dir=args.input,
                out_file=args.output,
                num_workers=args.num_workers,
                return_type=args.return_type,
                batch_size=args.batch_size
            )
        else:
            print("Unknown command: {}".format(args.command))



import torch
import torch.nn as nn
import torch.nn.functional as F

class SEBasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, reduction=8):
        super(SEBasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.se = SELayer(planes, reduction)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.relu(out)
        out = self.bn1(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.se(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)
        return out

class SELayer(nn.Module):
    def __init__(self, channel, reduction=8):
        super(SELayer, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
                nn.Linear(channel, channel // reduction),
                nn.ReLU(inplace=True),
                nn.Linear(channel // reduction, channel),
                nn.Sigmoid()
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y

class audioEncoder(nn.Module):
    def __init__(self, layers, num_filters, **kwargs):
        super(audioEncoder, self).__init__()
        block = SEBasicBlock
        self.inplanes   = num_filters[0]

        self.conv1 = nn.Conv2d(1, num_filters[0] , kernel_size=7, stride=(2, 1), padding=3,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(num_filters[0])
        self.relu = nn.ReLU(inplace=True)

        self.layer1 = self._make_layer(block, num_filters[0], layers[0])
        self.layer2 = self._make_layer(block, num_filters[1], layers[1], stride=(2, 2))
        self.layer3 = self._make_layer(block, num_filters[2], layers[2], stride=(2, 2))
        self.layer4 = self._make_layer(block, num_filters[3], layers[3], stride=(1, 1))
        out_dim = num_filters[3] * block.expansion

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = torch.mean(x, dim=2, keepdim=True)
        x = x.view((x.size()[0], x.size()[1], -1))
        x = x.transpose(1, 2)

        return x


from ..baseExtractor import baseTextExtractor
import numpy as np


class gloveExtractor(baseTextExtractor):
    """
    Text feature extractor using GLOVE
    Ref: https://huggingface.co/docs/transformers/model_doc/bert
    Pretrained models: https://huggingface.co/models
    """
    def __init__(self, config, logger):
        try:
            logger.info("Initializing BERT text feature extractor.")
            super().__init__(config, logger)
            self.embedding, self.dim = self.get_glove_embeddings(config['embedding_file'])
        except Exception as e:
            logger.error("Failed to initialize gloveExtractor.")
            raise e

    def get_glove_embeddings(self, emb_file):
        res = {}
        with open(emb_file, "r") as f:
            while data := f.readline():
                data = data.split(' ')
                res[data[0]] = np.asarray(data[1:])
        return res, len(res['the'])
    
    def extract(self, text):
        try:
            text_list = text.split(' ')
            result = []
            for word in text_list:
                if word.lower() in self.embedding:
                    result.append(self.embedding[word.lower()])
                else:
                    result.append(np.zeros(self.dim))
            return np.asarray(result)
        except Exception as e:
            self.logger.error(f"Failed to extract text features with Glove for '{text}'.")
            raise e
    
    def tokenize(self, text):
        return None


import json
import logging
import multiprocessing
import os
import os.path as osp
import pickle
import shutil
import time
from glob import glob
from logging.handlers import RotatingFileHandler
from pathlib import Path

import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm

from .ASD import run_ASD
from .dataloader import FET_Dataset
from .extractors import *
from .utils import *


class FeatureExtractionTool(object):
    """
    Feature Extraction Tool for Multimodal Sentiment Analysis tasks.

    Parameters:
        config: 
            Python dictionary or path to a JSON file.
        dataset_root_dir: 
            Path to dataset root directory. Used when extracting dataset features.
        tmp_dir: 
            Temporary directory path. Default: '~/.MSA-FET/tmp'.
        log_dir: 
            Log directory path. Default: '~/.MSA-FET/log'.
        verbose: 
            Verbose level of stdout. 0 for error, 1 for info, 2 for debug. Default: 1.

    TODOs:
        1. Add csv/dataframe output format.
        2. Support specifying existing feature files, modify only some of the modalities.
    """

    def __init__(
        self,
        config,
        dataset_root_dir=None,
        tmp_dir=osp.join(Path.home(), '.MMSA-FET/tmp'),
        log_dir=osp.join(Path.home(), '.MMSA-FET/log'),
        verbose=1
    ):
        if type(config) == dict:
            self.config = config
        elif type(config) == str:
            if osp.isfile(config):
                with open(config, 'r') as f:
                    self.config = json.load(f)
            elif osp.isfile(name := osp.join(osp.dirname(__file__), 'example_configs', config + '.json')):
                with open(name, 'r') as f:
                    self.config = json.load(f)
            elif osp.isfile(name := osp.join(osp.dirname(__file__), 'example_configs', config)):
                with open(name, 'r') as f:
                    self.config = json.load(f)
            else:
                raise ValueError(f"Config file {config} does not exist.")
        else:
            raise ValueError("Invalid config type.")
        self.tmp_dir = tmp_dir
        self.log_dir = log_dir
        self.dataset_root_dir = dataset_root_dir
        self.verbose = verbose

        if not osp.isdir(self.tmp_dir):
            Path(self.tmp_dir).mkdir(parents=True, exist_ok=True)
        if not osp.isdir(self.log_dir):
            Path(self.log_dir).mkdir(parents=True, exist_ok=True)
            
        self.logger = logging.getLogger("MMSA-FET")
        if self.verbose == 1:
            self.__set_logger(logging.INFO)
        elif self.verbose == 0:
            self.__set_logger(logging.ERROR)
        elif self.verbose == 2:
            self.__set_logger(logging.DEBUG)
        else:
            raise ValueError(f"Invalid verbose level '{self.verbose}'.")
        
        self.logger.info("")
        self.logger.info("========================== MMSA-FET Started ==========================")
        self.logger.info(f"Temporary directory: {self.tmp_dir}")
        self.logger.info(f"Log file: '{osp.join(self.log_dir, 'MMSA-FET.log')}'")
        # self.logger.info(f"Config file: '{self.config_file}'")
        self.logger.info(f"Config: {self.config}")

        self.video_extractor, self.audio_extractor, self.text_extractor = None, None, None
        

    def __set_logger(self, stream_level):
        self.logger.setLevel(logging.DEBUG)

        fh = RotatingFileHandler(osp.join(self.log_dir, 'MSA-FET.log'), maxBytes=2e7, backupCount=5)
        fh_formatter = logging.Formatter('%(asctime)s - %(name)s [%(levelname)s] - %(message)s')
        fh.setFormatter(fh_formatter)
        fh.setLevel(logging.DEBUG)
        self.logger.addHandler(fh)

        ch = logging.StreamHandler()
        ch.setLevel(stream_level)
        ch_formatter = logging.Formatter('%(name)s - %(message)s')
        ch.setFormatter(ch_formatter)
        self.logger.addHandler(ch)

    def __init_extractors(self):
        if 'audio' in self.config and self.audio_extractor is None:
            # self.logger.info(f"Initializing audio feature extractor...")
            audio_cfg = self.config['audio']
            extractor_name = audio_cfg['tool']
            self.audio_extractor = AUDIO_EXTRACTOR_MAP[extractor_name](audio_cfg, self.logger)
        if 'video' in self.config and self.video_extractor is None:
            # self.logger.info(f"Initializing video feature extractor...")
            video_cfg = self.config['video']
            extractor_name = video_cfg['tool']
            self.video_extractor = VIDEO_EXTRACTOR_MAP[extractor_name](video_cfg, self.logger)
        if 'text' in self.config and self.text_extractor is None:
            # self.logger.info(f"Initializing text feature extractor...")
            text_cfg = self.config['text']
            extractor_name = text_cfg['model']
            self.text_extractor = TEXT_EXTRACTOR_MAP[extractor_name](text_cfg, self.logger)
    
    def __audio_extract_single(self, in_file, keep_tmp_file=False):
        # extract audio from video file
        # extension = get_codec_name(in_file, 'audio')
        tmp_audio_file = osp.join(self.tmp_dir, 'tmp_audio.wav')
        ffmpeg_extract(in_file, tmp_audio_file, mode='audio')
        
        # extract audio features
        audio_result = self.audio_extractor.extract(tmp_audio_file)
        # delete tmp audio file
        if not keep_tmp_file:
            os.remove(tmp_audio_file)
        return audio_result

    def __video_extract_single(self, in_file, keep_tmp_file=False):
        # extract images from video
        fps = self.config['video']['fps']
        if 'multiFace' in self.config['video'] and self.config['video']['multiFace']['enable'] == True:
            # enable Active Speaker Detection
            run_ASD(in_file, self.tmp_dir, fps, self.config['video']['multiFace'])
        else:
            ffmpeg_extract(in_file, self.tmp_dir, mode='image', fps=fps)

        # extract video features
        name = 'video_' + Path(in_file).stem
        video_result = self.video_extractor.extract(self.tmp_dir, name, tool_output=self.verbose>0)
        # delete tmp images
        if not keep_tmp_file:
            for image_path in glob(osp.join(self.tmp_dir, '*.bmp')):
                os.remove(image_path)
            for image_path in glob(osp.join(self.tmp_dir, '*.jpg')):
                os.remove(image_path)
        return video_result

    def __text_extract_single(self, in_file):
        text = self.text_extractor.load_text(in_file)
        text_result = self.text_extractor.extract(text)
        # text_tokens = self.text_extractor.tokenize(text)
        return text_result

    def __read_label_file(self, dataset_name, dataset_root_dir, dataset_dir):
        # Locate and read label.csv file
        assert dataset_name is not None or dataset_dir is not None, "Either 'dataset_name' or 'dataset_dir' must be specified."
        if dataset_dir: # Use dataset_dir
            dataset_dir = osp.normpath(dataset_dir) # normalize path, remove trailing slash so that osp.basename function works correctly
            dataset_name = osp.basename(dataset_dir)
            if not osp.exists(dataset_dir):
                raise RuntimeError(f"Dataset directory '{self.dataset_dir}' does not exist.")
            if not osp.exists(osp.join(dataset_dir, 'label.csv')):
                raise RuntimeError(f"Label file '{dataset_dir}/label.csv' does not exist.")
            label_df = pd.read_csv(
                osp.join(dataset_dir, 'label.csv'),
                dtype={'clip_id': str, 'video_id': str, 'text': str}
            )
            return label_df, dataset_dir, dataset_name, None
        else: # Use dataset_name
            dataset_root_dir = dataset_root_dir if dataset_root_dir is not None else self.dataset_root_dir
            if dataset_root_dir is None:
                raise ValueError("Dataset root directory is not specified.")
            if not osp.exists(dataset_root_dir):
                raise RuntimeError(f"Dataset root directory '{dataset_root_dir}' does not exist.")
            try: # Try to locate label.csv according to global dataset config file
                with open(osp.join(self.dataset_root_dir, 'config.json'), 'r') as f:
                    dataset_config_all = json.load(f)
                dataset_config = dataset_config_all[dataset_name]
                label_file = osp.join(self.dataset_root_dir, dataset_config['label_path'])
            except: # If failed, try to locate label.csv using joined path
                label_file = osp.join(dataset_root_dir, dataset_name, 'label.csv')
            if not osp.exists(label_file):
                raise RuntimeError(f"Label file '{label_file}' does not exist.")
            label_df = pd.read_csv(
                label_file,
                dtype={'clip_id': str, 'video_id': str, 'text': str}
            )
            return label_df, osp.dirname(label_file), dataset_name, dataset_config

    def __padding(self, feature, MAX_LEN, value='zero', location='end'):
        """
        Parameters:
            mode: 
                zero: padding with 0
                norm: padding with normal distribution
            location: start / end
        """
        assert value in ['zero', 'norm'], "Padding value must be 'zero' or 'norm'"
        assert location in ['start', 'end'], "Padding location must be 'start' or 'end'"

        length = feature.shape[0]
        if length >= MAX_LEN:
            return feature[:MAX_LEN, :]
        
        if value == "zero":
            pad = np.zeros([MAX_LEN - length, feature.shape[-1]])
        elif value == "normal":
            mean, std = feature.mean(), feature.std()
            pad = np.random.normal(mean, std, (MAX_LEN-length, feature.shape[1]))

        feature = np.concatenate((pad, feature), axis=0) if(location == "start") else \
                  np.concatenate((feature, pad), axis=0)
        return feature

    def __paddingSequence(self, sequences):
        """
        Pad features to the same length according to the mean length of the features.
        """
        feature_dim = sequences[0].shape[-1]
        lengths = [s.shape[0] for s in sequences]
        # use (mean + 3 * std) as the max length
        final_length = int(np.mean(lengths) + 3 * np.std(lengths))
        final_sequence = np.zeros([len(sequences), final_length, feature_dim])
        for i, s in enumerate(sequences):
            if len(s) != 0:
                final_sequence[i] = self.__padding(s, final_length)
        return final_sequence, final_length

    def __collate_fn(self, batch):
        res = None
        for b in batch: # need to iterate through batch in case the first sample is bad(None)
            if b is not None:
                res = {k: [] for k in batch[0].keys()} # initialize res
                break
        if res is None: # if all samples in this batch are bad(None), return None
            return None
        for b in batch:
            if b is None: # if one sample is bad(None), skip it
                continue
            for k, v in b.items():
                res[k].append(v)
        return res

    def __save_result(self, result, out_file):
        if osp.exists(out_file):
            out_file_alt = osp.splitext(out_file)[0] + '_' + str(int(time.time())) + '.pkl'
            self.logger.warning(f"Output file '{out_file}' already exists. Saving to '{out_file_alt}' instead.")
            out_file = out_file_alt
        Path(out_file).parent.mkdir(parents=True, exist_ok=True)
        with open(out_file, 'wb') as f:
            pickle.dump(result, f)
        self.logger.info(f"Feature file saved: '{out_file}'.")

    def __remove_tmp_folder(self, tmp_dir):
        if osp.exists(tmp_dir):
            shutil.rmtree(tmp_dir)

    def run_single(self, in_file, out_file=None, text_file=None, return_type='np'):
        """
        Extract features from single file.

        Parameters:
            in_file: path to input video file.
            return_type: 'pt' for pytorch tensor, 'np' for numpy array. Default: 'np'.
            out_file (optional): path to output file.
            text_file (optional): path to text file.
        
        Returns:
            final_result: dictionary of extracted features.
        """
        try:
            self.__init_extractors()
            self.logger.info(f"Extracting features from '{in_file}'.")
            final_result = {}
            if 'audio' in self.config:
                audio_result = self.__audio_extract_single(in_file, keep_tmp_file=True)
            if 'video' in self.config:
                video_result = self.__video_extract_single(in_file)
            if 'text' in self.config:
                if text_file is None:
                    raise ValueError("Text file is not specified.")
                text_result = self.__text_extract_single(text_file)
            # combine audio and video features
            if return_type == 'pt':
                if 'audio' in self.config:
                    final_result['audio'] = torch.from_numpy(audio_result)
                if 'video' in self.config:
                    final_result['video'] = torch.from_numpy(video_result)
                if 'text' in self.config:
                    final_result['text'] = torch.from_numpy(text_result)
            elif return_type == 'np':
                if 'audio' in self.config:
                    final_result['audio'] = audio_result
                if 'video' in self.config:
                    final_result['video'] = video_result
                if 'text' in self.config:
                    final_result['text'] = text_result
            else:
                raise ValueError(f"Invalid return type '{return_type}'.")
            # save result
            if out_file:
                self.__save_result(final_result, out_file)
            return final_result
        except Exception as e:
            self.logger.exception("An Error Occured:")
            self.logger.debug("Removing temporary files.")
            self.__remove_tmp_folder(self.tmp_dir)
            raise e

    def run_dataset(self, dataset_name=None, dataset_root_dir=None, dataset_dir=None, out_file=None, return_type='np', num_workers=4, batch_size=32, skip_bad_data=True, progress_q=None, task_id=None):
        """
        Extract features from dataset and save in MMSA compatible format.

        Parameters:
            dataset_name: name of dataset. Either 'dataset_name' or 'dataset_dir' must be specified.
            dataset_root_dir: root directory of dataset. If specified, will override 'dataset_root_dir' set when initializing MSA-FET.
            dataset_dir: Path to dataset directory. If specified, will override 'dataset_name'. Either 'dataset_name' or 'dataset_dir' must be specified.
            out_file: output feature file. If not specified, features will be saved under the dataset directory with the name 'feature.pkl'.
            return_type: 'pt' for pytorch tensor, 'np' for numpy array. Default: 'np'.
            num_workers: number of workers for parallel processing. Default: 4.
            batch_size: batch size for parallel processing. Default: 32.
            skip_bad_data: skip bad data when loading dataset. Default: True.
            progress_q: multiprocessing queue for progress reporting with M-SENA.
            task_id: task id for M-SENA.
        """
        # TODO: add database operation for M-SENA
        try:
            self.label_df, self.dataset_dir, self.dataset_name, self.dataset_config = \
                self.__read_label_file(dataset_name, dataset_root_dir, dataset_dir)
            
            self.logger.info(f"Extracting features from '{self.dataset_name}' dataset.")
            self.logger.info(f"Dataset directory: '{self.dataset_dir}'")

            self.report = None
            if type(progress_q) == multiprocessing.queues.Queue and task_id is not None:
                self.report = {'task_id': task_id, 'msg': 'Preparing', 'processed': 0, 'total': 0}
                progress_q.put(self.report)

            data = {
                "id": [], 
                "audio": [],
                "vision": [],
                "raw_text": [],
                "text": [],
                "text_bert": [],
                "audio_lengths": [],
                "vision_lengths": [],
                "annotations": [],
                # "classification_labels": [],  # no longer supported by MMSA
                "regression_labels": [],
                'regression_labels_A': [],
                'regression_labels_V': [],
                'regression_labels_T': [],
                "mode": []
            }

            dataloader = DataLoader(
                FET_Dataset(
                    self.label_df, self.dataset_dir, self.dataset_name,
                    self.config, self.dataset_config, self.tmp_dir, ignore_error=skip_bad_data
                ),
                batch_size=batch_size,
                num_workers=num_workers,
                shuffle=False,
                collate_fn=self.__collate_fn,
                # multiprocessing_context='spawn'
                # Using 'spawn' instead of 'fork' lead to more errors
                # Pytorch dataloader currently does not support cuda multiprocessing
                # Watch https://github.com/pytorch/pytorch/issues/41292 for updates
                # Currently only cpu is supported for dataset feature extraction
            )
            if self.report is not None:
                self.report['msg'] = 'Processing'
                self.report['total'] = len(dataloader)
                progress_q.put(self.report)
            for i, batch_data in enumerate(tqdm(dataloader)):
                if batch_data is None: # if all samples in this batch are bad(None), skip the batch
                    continue
                for k, v in batch_data.items():
                    data[k].extend(v)
                if self.report is not None:
                    self.report['processed'] = i + 1
                    progress_q.put(self.report)
            if self.report is not None:
                self.report['msg'] = 'Finalizing'
                progress_q.put(self.report)
            
            # remove unimodal labels if not exist
            for key in ['regression_labels_A', 'regression_labels_V', 'regression_labels_T']:
                if np.isnan(np.sum(data[key])):
                    data.pop(key)
            # remove empty features
            for key in ['audio', 'vision', 'text', 'text_bert', 'audio_lengths', 'vision_lengths']:
                if len(data[key]) == 0:
                    data.pop(key)
            # padding features
            for item in ['audio', 'vision', 'text', 'text_bert']:
                if item in data:
                    data[item], final_length = self.__paddingSequence(data[item])
                    if f"{item}_lengths" in data:
                        for i, length in enumerate(data[f"{item}_lengths"]):
                            if length > final_length:
                                data[f"{item}_lengths"][i] = final_length
            # transpose text_bert
            if 'text_bert' in data:
                data['text_bert'] = data['text_bert'].transpose(0, 2, 1)
            # repack features
            idx_dict = {
                mode + '_index': [i for i, v in enumerate(data['mode']) if v == mode]
                for mode in ['train', 'valid', 'test']
            }
            data.pop('mode')
            final_data = {k: {} for k in ['train', 'valid', 'test']}
            for mode in ['train', 'valid', 'test']:
                indexes = idx_dict[mode + '_index']
                for item in data.keys():
                    if isinstance(data[item], list):
                        final_data[mode][item] = np.array([data[item][v] for v in indexes])
                    else:
                        final_data[mode][item] = data[item][indexes]
            data = final_data
            # convert labels to numpy array

            # convert to pytorch tensors
            if return_type == 'pt':
                for mode in data.keys():
                    for key in ['audio', 'vision', 'text', 'text_bert']:
                        if key in data[mode]:
                            data[mode][key] = torch.from_numpy(data[mode][key])
            # save result
            if out_file is None:
                out_file = osp.join(self.dataset_dir, 'feature.pkl')
            self.__save_result(data, out_file)
            self.logger.info(f"Feature extraction complete!")
            if self.report is not None:
                self.report['msg'] = 'Finished'
                progress_q.put(self.report)
            return data
        except KeyboardInterrupt:
            self.logger.info("User aborted feature extraction!")
            self.__remove_tmp_folder(self.tmp_dir)
            if self.report is not None:
                self.report['msg'] = 'Terminated'
                progress_q.put(self.report)
        except Exception as e:
            self.logger.exception("An Error Occured:")
            self.logger.info("Removing temporary files.")
            self.__remove_tmp_folder(self.tmp_dir)
            if self.report is not None:
                self.report['msg'] = 'Error'
                progress_q.put(self.report)
            raise e




from ..baseExtractor import baseAudioExtractor
import opensmile

class opensmileExtractor(baseAudioExtractor):
    """
    Audio feature extractor using openSMILE. 
    Ref: https://github.com/audeering/opensmile-python
    """
    def __init__(self, config, logger):
        try:
            logger.info("Initializing openSMILE audio feature extractor")
            super().__init__(config, logger)
            self.args = self.config['args']
            self.extractor = opensmile.Smile(
                feature_set=eval(f"opensmile.FeatureSet.{self.args['feature_set']}"),
                feature_level=eval(f"opensmile.FeatureLevel.{self.args['feature_level']}"),
                num_workers=None,
            )
        except Exception as e:
            self.logger.error("Failed to initialize opensmileExtractor.")
            raise e

    def extract(self, file):
        """
        Function:
            Extract features from audio file using openSMILE.

        Parameters:
            file: path to audio file

        Returns:
            audio_result: extracted audio features in numpy array
        """
        try:
            y, sr = self.load_audio(file)
            audio_result = self.extractor.process_signal(
                signal = y, 
                sampling_rate = sr, 
                start = self.args['start'],
                end = self.args['end']
            )
            audio_result = self.extractor.to_numpy(audio_result).squeeze(0).transpose()
            return audio_result
        except Exception as e:
            self.logger.error(f"Failed to extract audio features with openSMILE from {file}")
            raise e